"""Miscellaneous commands."""

__all__ = [
    "comment",
    "user_prompt",
    "set_variable",
    "start_timer",
    "wait_timer",
    "facts",
]


def comment(msg: str):
    """Write comment."""
    return f'B;Comment("{msg}");'.encode()


def user_prompt(msg: str, sound: bool = False, timeout: int = -1):
    """Open user prompt."""
    return f'B;UserPrompt("{msg}",{sound:d},{timeout});'.encode()


def set_variable(name: str, value: int):
    """Set a variable."""
    return f'B;Variable({name},"{value}",0,"",0,1.000000,10.000000,0,2,0,0);'.encode()


def start_timer(index: int):
    """Start timer."""
    return f'B;StartTimer("{index}");'.encode()


def wait_timer(index: int, time: int):
    """Wait for timer."""
    return f'B;WaitTimer("{index}","{time}");'.encode()


def facts(
    device: str,
    command: str,
    parameter: str,
    needs_labware: bool,
    allowed_labware: str,
):
    """General FACTS command"""
    return (
        "B;FACTS("
        f'"{device}",'
        f'"{command}",'
        f'"{parameter}",'
        f'"{needs_labware:d}",'
        f'"{allowed_labware}",'
        ");"
    ).encode()



"""Pipetting functions."""

import numpy as np

from numpy.typing import ArrayLike

from .base import GridSite, Labware
from .labware import labwares
from .utils import bin_to_dec, volumes_to_string

__all__ = [
    "aspirate",
    "dispense",
    "mix",
    "wash",
    "move_liha",
]


def aspirate(
    grid_site: GridSite,
    column: int,
    column_mask: ArrayLike,
    volumes: ArrayLike | int | float,
    liquid_class: str,
    spacing: int = 1,
    labware: Labware | str = "greiner96",
):
    """Advanced aspirate command.

    Parameters
    ----------
    grid_site : GridSite
    column : int
    column_mask : array_like
        Boolean mask to indicate which wells to aspirate from
    volumes : array_like | int | float
        Array showing which volumes each tip aspirates; a number is given, all
        tips will aspirate that volume.
    spacing : int, optional
        Spacing of the tips with respect to the column mask
    labware : Labware | str, optional
        Labware specification to be used. (Default: Greiner 96 Well Plate)

    """

    return _liha_command(
        cmd="Aspirate",
        grid_site=grid_site,
        column=column,
        column_mask=column_mask,
        volumes=volumes,
        liquid_class=liquid_class,
        spacing=spacing,
        labware=labware,
    )


def dispense(
    grid_site: GridSite,
    column: int,
    column_mask: ArrayLike,
    volumes: ArrayLike | int | float,
    liquid_class: str,
    spacing: int = 1,
    labware: Labware | str = "greiner96",
):
    """Advanced dispense command."""

    return _liha_command(
        cmd="Dispense",
        grid_site=grid_site,
        column=column,
        column_mask=column_mask,
        volumes=volumes,
        liquid_class=liquid_class,
        spacing=spacing,
        labware=labware,
    )


def mix(
    grid_site: GridSite,
    column: int,
    column_mask: ArrayLike,
    volumes: ArrayLike | int | float,
    liquid_class: str,
    spacing: int = 1,
    labware: Labware | str = "greiner96",
):
    """Advanced mix command."""

    return _liha_command(
        cmd="Mix",
        grid_site=grid_site,
        column=column,
        column_mask=column_mask,
        volumes=volumes,
        liquid_class=liquid_class,
        spacing=spacing,
        labware=labware,
    )


def wash(waste_volume: int | float, cleaner_volume: int | float, station: int):
    """Wash tips command."""
    return (
        "B;Wash("
        "255,"
        f"{station},1,{station},0,"
        f'"{waste_volume}",'
        f"500,"
        f'"{cleaner_volume}",'
        "0,10,70,30,0,0,1000,0);"
    ).encode()


def move_liha(
    grid_site: GridSite,
    column: int,
    positions: ArrayLike = None,
    spacing: int = 1,
    labware: Labware | str = "greiner96",
    local: bool = False,
    z_pos: int = 0,
    speed: int = 10,
):
    """Move LiHa to grid site."""

    if isinstance(labware, str):
        labware = labwares[labware]

    if positions is None:
        positions = np.array([True] * 8)

    if column not in range(1, labware.cols + 1):
        raise IndexError(f"Plate column out of bounds: {column=}")

    command = (
        (
            "B;MoveLiha("
            "255,"
            f"{grid_site.grid},"
            f"{grid_site.site},"
            f'{spacing},"'
        ).encode()
        + _well_select(positions, column, labware.rows, labware.cols)
        + (f'",{local:#d},{z_pos},0,{speed},0,0);').encode()
    )

    return command


def _encode_well_select(column_mask, offset):
    encoder = np.zeros(offset + len(column_mask), dtype=np.int8)
    for i, volume in enumerate(column_mask):
        encoder[offset + i] = 1 if volume > 0 else 0
    splits = [encoder[(7 * i) : (7 * (i + 1))] for i in range((len(encoder) + 6) // 7)]
    return [2 ** np.arange(len(split)) * split for split in splits]


def _well_select(column_mask, column, nrows, ncols, **kwargs):
    """Generate well select string for volumes."""
    well = (column - 1) * nrows
    encoders = _encode_well_select(column_mask, well % 7)
    n_chars = (nrows * ncols + 6) // 7
    sequence = [
        f"{ncols:02x}".encode(),
        f"{nrows:02x}".encode(),
        b"0" * int(well / 7),
        *[
            int(sum(encoder) + 48).to_bytes(1, "little") if sum(encoder) > 0 else b"0"
            for encoder in encoders
        ],
        b"0" * (n_chars - len(encoders) - int(well / 7)),
    ]
    return b"".join(sequence)


def _liha_command(
    cmd: str,
    grid_site: GridSite,
    column: int,
    column_mask: ArrayLike,
    volumes: ArrayLike | int | float,
    liquid_class: str,
    spacing: int = 1,
    labware: Labware | str = "greiner96",
):
    if isinstance(volumes, int | float):
        volumes = np.array(column_mask * volumes)

    if isinstance(labware, str):
        labware = labwares[labware]

    if column not in range(1, labware.cols + 1):
        raise IndexError(f"Plate column out of bounds: {column=}")

    command = (
        (
            f"B;{cmd}("
            f"{bin_to_dec(volumes)},"
            f'"{liquid_class}",'
            f"{volumes_to_string(volumes)},"
            f"{grid_site.grid},"
            f"{grid_site.site},"
            f'{spacing},"'
        ).encode()
        + _well_select(column_mask, column, labware.rows, labware.cols)
        + b'",0,0);'
    )
    return command



"""Base classes."""
from dataclasses import dataclass

from dataclasses_json import DataClassJsonMixin

from .errors import GridSiteException

__all__ = [
    "GridSite",
    "GridStack",
    "GridQueue",
    "Labware",
]


class _DataClassCached(DataClassJsonMixin):
    """Wrapper class to add file writing to dataclass json serde."""

    def write_to_file(self, path):
        """Write dataclass to file."""
        with open(path, "w", encoding="utf8") as file_descriptor:
            file_descriptor.write(self.to_json())

    @classmethod
    def load_from_file(cls, path):
        """Load dataclass from file."""
        with open(path, "r", encoding="utf8") as file_descriptor:
            return cls.from_json(file_descriptor.read())


@dataclass(frozen=True)
class GridSite:
    """Specify a grid site."""

    grid: int
    site: int
    carrier: str


@dataclass
class GridStack(_DataClassCached):
    """Specify a stash."""

    grid: int
    capacity: int
    carrier: str
    size: int = 0

    def push(self) -> GridSite:
        """Add element to stash."""
        if self.size >= self.capacity:
            raise GridSiteException("Can not push element to full GridStash.")
        self.size += 1
        return GridSite(self.grid, self.size - 1, self.carrier)

    def pop(self) -> GridSite:
        """Remove element from stash."""
        if self.size <= 0:
            raise GridSiteException("Can not pop element from empty GridStash.")
        self.size -= 1
        return GridSite(self.grid, self.size, self.carrier)


@dataclass
class GridQueue(_DataClassCached):
    """Round robin queue."""

    grid: int
    capacity: int
    carrier: str
    first: int = 0
    last: int = 0

    def push(self) -> GridSite:
        """Add element to queue."""
        if self.size >= self.capacity:
            raise GridSiteException("Can not push element to full GridQueue.")
        self.last = (self.last + 1) % (self.capacity + 1)
        return GridSite(self.grid, self.last - 1, self.carrier)

    def pop(self) -> GridSite:
        """Get element from queue."""
        if self.size <= 0:
            raise GridSiteException("Can not pop element from full GridQueue.")
        self.first = (self.first + 1) % (self.capacity + 1)
        return GridSite(self.grid, self.first - 1, self.carrier)

    @property
    def size(self) -> int:
        """Number of elements in the queue."""
        size = self.last - self.first
        if size < 0:
            size = size + self.capacity
        return size


@dataclass(frozen=True)
class Labware:
    """Labware specification."""

    name: str
    rows: int
    cols: int
    spacing: int = 1



"""Labware dimensions."""

from .base import Labware

labwares = {
    "trough100": Labware("Trough 100+25ml", 8, 1),
    "greiner96": Labware("96 Well Greiner SC", 8, 12),
    "greiner384": Labware("384 Well Plate Greiner", 16, 24, 2),
}



"""Pipetting unit tests."""

from itertools import product

import pytest
import numpy as np

from pypetting import (
    GridSite,
    aspirate,
    dispense,
    mix,
    wash,
    move_liha,
)

from pypetting.liha import _well_select

FIRST = np.array([True] + 7 * [False])
COL96 = np.array(8 * [True])
grid_site = GridSite(40, 0, "MP 3Pos")

LIHA_PIPETTING_ARGS = (
    b'(1,"A","1.0","0","0","0","0","0","0","0","0","0","0","0",40,0,1,"010810",0,0);'
)


def test_liha_aspirate():
    """Test LiHa aspirate."""
    assert (
        aspirate(grid_site, 1, FIRST, 1 * FIRST, "A", labware="trough100")
        == b"B;Aspirate" + LIHA_PIPETTING_ARGS
    )
    assert (
        aspirate(grid_site, 1, FIRST, 1, "A", labware="trough100")
        == b"B;Aspirate" + LIHA_PIPETTING_ARGS
    )
    with pytest.raises(IndexError):
        aspirate(grid_site, 0, FIRST, 1, "A", labware="trough100")
    with pytest.raises(IndexError):
        aspirate(grid_site, 2, FIRST, 1, "A", labware="trough100")


def test_liha_dispense():
    """Test Liha dispense."""
    assert (
        dispense(grid_site, 1, FIRST, 1 * FIRST, "A", labware="trough100")
        == b"B;Dispense" + LIHA_PIPETTING_ARGS
    )
    assert (
        dispense(grid_site, 1, FIRST, 1, "A", labware="trough100")
        == b"B;Dispense" + LIHA_PIPETTING_ARGS
    )
    with pytest.raises(IndexError):
        dispense(grid_site, 0, FIRST, 1, "A", labware="trough100")
    with pytest.raises(IndexError):
        dispense(grid_site, 2, FIRST, 1, "A", labware="trough100")


def test_liha_mix():
    """Test Liha mix."""
    assert (
        mix(grid_site, 1, FIRST, 1 * FIRST, "A", labware="trough100")
        == b"B;Mix" + LIHA_PIPETTING_ARGS
    )
    assert (
        mix(grid_site, 1, FIRST, 1, "A", labware="trough100")
        == b"B;Mix" + LIHA_PIPETTING_ARGS
    )
    with pytest.raises(IndexError):
        mix(grid_site, 0, FIRST, 1, "A", labware="trough100")
    with pytest.raises(IndexError):
        mix(grid_site, 2, FIRST, 1, "A", labware="trough100")


def test_wash():
    """Test LiHa wash."""
    assert (
        wash(waste_volume=1, cleaner_volume=0.1, station=10)
        == b'B;Wash(255,10,1,10,0,"1",500,"0.1",0,10,70,30,0,0,1000,0);'
    )


def test_move_liha():
    """Test LiHa move."""
    assert (
        move_liha(grid_site, 1, labware="trough100")
        == b'B;MoveLiha(255,40,0,1,"0108\xaf1",0,0,0,10,0,0);'
    )


def test_well_select_minimal():
    """Test well select string with minimal length."""
    for mask in product([0, 1], repeat=7):
        value = 48
        for idx, val in enumerate(mask):
            value += val * 2 ** idx
        assert _well_select(mask, 1, 7, 1) == b"0107" + value.to_bytes(1, "big")



"""Infinite Reader Commands."""

from .misc import facts

__all__ = [
    "open_infinite_reader",
    "close_infinite_reader",
    "measure_infinite_reader",
]


def open_infinite_reader():
    """Open infinite reader."""
    return facts("ReaderNETwork", "ReaderNETwork_Open", "", 0, "")


def close_infinite_reader():
    """Close infinite reader."""
    return facts("ReaderNETwork", "ReaderNETwork_Close", "", 0, "")


def measure_infinite_reader(path: str, settings_path: str):
    """Measure infinite reader."""
    with open(settings_path, encoding="utf8") as settings_file:
        settings = settings_file.read()

    settings_string = (
        settings.replace('"', "&quote;")
        .replace("=", "&equal;")
        .replace("  ", "")
        .replace("\n", " ")
        .replace("> <", "><")
        .replace(" />", "/>")
    )

    return facts(
        "ReaderNETwork",
        "ReaderNETwork_Measure",
        f"{path}|{settings_string}",
        0,
        "",
    )



"""Module for custom exceptions."""

__all__ = ["GridSiteException"]


class PypettingException(Exception):
    """Base class for pypetting exceptions."""


class GridSiteException(PypettingException):
    """Exceptions related to container elements on the grid."""

    def __init__(self, msg):
        self.msg = msg
        super().__init__()



"""Teshake module tests."""

from pypetting.teshake import teshake_shake


def test_shake():
    """Test shake command."""
    assert teshake_shake() == (
        b'B;FACTS("Shaker","Shaker_Init","","0","",);\n'
        b'B;FACTS("Shaker","Shaker_SetFrequency","1000","0","",);\n'
        b'B;FACTS("Shaker","Shaker_Start","1","0","",);\n'
        b'B;StartTimer("40");\n'
        b'B;WaitTimer("40","10");\n'
        b'B;FACTS("Shaker","Shaker_Stop","","0","",);'
    )



"""Utils."""

import numpy as np


def bin_to_dec(array):
    """Convert binary vector to decimal."""
    return sum((array > 0) * 2 ** np.arange(len(array)))


def _format_number(num):
    return '"0"' if num == 0 else f'"{str(num)}"'


def volumes_to_string(volumes):
    """Convert volumes vector to string."""
    if len(volumes) < 12:
        volumes = np.append(volumes, np.zeros(12 - len(volumes)))
    return ",".join(map(_format_number, volumes))



"""Transfer labware."""

from .base import GridSite, Labware

from .labware import labwares

__all__ = ["transfer_labware"]


def transfer_labware(
    src: GridSite,
    dest: GridSite,
    labware: Labware | str,
    lid: GridSite = None,
    cover: bool = False,
    back_home: bool = False,
):
    """Transfer labware with RoMa."""
    if isinstance(labware, str):
        labware = labwares[labware]

    return (
        "B;Transfer_Rack("
        f'"{src.grid}",'
        f'"{dest.grid}",'
        f"{back_home:#d},"
        f"{bool(lid):#d},"
        "1,0,"
        f"{not cover:#d},"
        f'"{lid.grid if lid is not None else ""}",'
        f'"{labware.name}",'
        '"Narrow",'
        '"","",'
        f'"{src.carrier}",'
        f'"{lid.carrier if lid is not None else ""}",'
        f'"{dest.carrier}",'
        f'"{src.site + 1}",'
        f'"{lid.site + 1 if lid is not None else ""}",'
        f'"{dest.site + 1}");'
    ).encode()



"""FACTS commands, specifically for StoreX incubator"""

from .misc import facts
from .labware import labwares

__all__ = [
    "read_all_barcodes",
    "read_plates_in_cartridge",
    "present_plate",
    "present_plate_by_bc",
    "return_plate",
    "return_plate_by_bc",
]


def read_all_barcodes():
    """Read all barcodes in StoreX"""
    return _facts_store_x("StoreX_ReadAllBarcodes", "", False, "")


def read_plates_in_cartridge(cartridge: int):
    """Read barcodes in cartridge"""
    return _facts_store_x("StoreX_ReadPlatesInCartridge", cartridge, False, "")


def present_plate(cartridge: int, position: int, labware: str):
    """Present plate"""
    return _facts_store_x(
        "StoreX_PresentPlate", f"{cartridge},{position}", True, labwares[labware].name
    )


def return_plate(cartridge: int, position: int):
    """Present plate"""
    return _facts_store_x("StoreX_ReturnPlate", f"{cartridge},{position}", False, "")


def present_plate_by_bc(bc_file_path, bc_file_line, labware):
    """Present plate by barcode from file"""
    return _facts_store_x(
        "StoreX_PresentPlateByBarcode",
        f"{bc_file_line},{bc_file_path}",
        1,
        labwares[labware].name,
    )


def return_plate_by_bc():
    """Return plate by barcode from file"""
    return _facts_store_x("StoreX_ReturnPlateByBarcode", "", False, "")


def _end_access():
    """End StoreX interactions"""
    return facts("StoreX", "StoreX_EndAccess", "", False, "")


def _facts_store_x(
    command: str, parameter: str, needs_labware: bool, allowed_labware: str
):
    """facts for storeX"""
    return (
        facts("StoreX", command, parameter, needs_labware, allowed_labware)
        + b"\n"
        + _end_access()
    )



"""Pypetting."""

from .base import *
from .errors import *
from .infinite import *
from .liha import *
from .mca import *
from .message import *
from .misc import *
from .roma import *
from .storex import *
from .teshake import *
from .write import *



"""Te-Shaker commands."""
from .misc import facts
from .misc import start_timer, wait_timer

__all__ = [
    "teshake_shake",
    "teshake_init",
    "teshake_start",
    "teshake_stop",
    "teshake_set_frequency",
]


def teshake_shake(frequency=1000, time=10, timer_id=40):
    """Shake with teshake."""
    return b"\n".join(
        [
            teshake_init(),
            teshake_set_frequency(frequency),
            teshake_start(),
            start_timer(timer_id),
            wait_timer(timer_id, time),
            teshake_stop(),
        ]
    )


def teshake_init():
    """Init teshake."""
    return facts("Shaker", "Shaker_Init", "", False, "")


def teshake_start(clockwise: bool = True):
    """Start teshake."""
    return facts("Shaker", "Shaker_Start", f"{1 + (not clockwise)}", False, "")


def teshake_stop():
    """Stop teshake."""
    return facts("Shaker", "Shaker_Stop", "", False, "")


def teshake_set_frequency(frequency=1000):
    """Start teshake."""
    return facts("Shaker", "Shaker_SetFrequency", f"{frequency}", False, "")



"""Send a message."""

import argparse
import requests

__all__ = ["send_message"]


def send_message(msg: str, url: str):
    """Send message from worklist."""
    return (f'B;Execute("pypetting-message-service {url} {msg}",0,"",2);').encode()


def message_service():
    """Message helper tool."""
    parser = argparse.ArgumentParser(description="Send messages to slack webhook.")
    parser.add_argument("url", type=str)
    parser.add_argument("msg", type=str, nargs="+")
    args = parser.parse_args()

    msg = " ".join(args.msg)
    requests.post(args.url, json={"text": msg})



"""Test base module."""

import pytest

from pypetting.base import GridStack, GridQueue
from pypetting.errors import GridSiteException


def test_grid_stack():
    """Test GridStack."""
    stack = GridStack(40, 2, "B")
    push1 = stack.push()
    push2 = stack.push()
    with pytest.raises(GridSiteException):
        stack.push()
    pop1 = stack.pop()
    pop2 = stack.pop()
    with pytest.raises(GridSiteException):
        stack.pop()
    assert push1 == pop2
    assert push2 == pop1


def test_grid_queue():
    """Test GridQueue"""
    queue = GridQueue(40, 2, "B")
    push1 = queue.push()
    push2 = queue.push()
    with pytest.raises(GridSiteException):
        queue.push()
    pop1 = queue.pop()
    pop2 = queue.pop()
    with pytest.raises(GridSiteException):
        queue.pop()
    assert push1 == pop1
    assert push2 == pop2



"""Misc module tests."""

from pypetting.misc import (
    comment,
    user_prompt,
    set_variable,
    start_timer,
    wait_timer,
    facts,
)


def test_comment():
    """Test comment."""
    assert comment("hello") == b'B;Comment("hello");'


def test_user_prompt():
    """Test user prompt."""
    assert user_prompt("hello") == b'B;UserPrompt("hello",0,-1);'


def test_set_variable():
    """Test set variable."""
    assert (
        set_variable("var_name", 100)
        == b'B;Variable(var_name,"100",0,"",0,1.000000,10.000000,0,2,0,0);'
    )


def test_start_timer():
    """Test start timer."""
    assert start_timer(1) == b'B;StartTimer("1");'


def test_wait_timer():
    """Test wait timer."""
    assert wait_timer(1, 100) == b'B;WaitTimer("1","100");'


def test_facts():
    """Test facts."""
    assert facts("D", "C", "0,0", False, "") == b'B;FACTS("D","C","0,0","0","",);'



"""Write work lists."""

__all__ = ["write_gwl"]


def write_gwl(path, gwl):
    """Write a gwl file."""
    with open(path, "wb") as file_descriptor:
        file_descriptor.write(b"C;INIT - gwl created by pypetting\n")
        file_descriptor.write(b"\n".join(gwl))



"""Tip handling."""

import numpy as np

from .base import GridSite, Labware
from .labware import labwares

__all__ = [
    "mca_aspirate",
    "mca_dispense",
    "mca_mix",
    "mca_get_tips",
    "mca_drop_tips",
]


def mca_aspirate(
    grid_site: GridSite,
    row: int,
    col: int,
    volume: int | float,
    liquid_class: str,
    labware: Labware | str = "greiner96",
):
    """Advanced aspirate command."""

    if isinstance(labware, str):
        labware = labwares[labware]

    command = (
        (
            "B;MCAAspirate("
            f'"{liquid_class}",'
            f'"{volume:g}",'
            f"{grid_site.grid},"
            f'{grid_site.site},"'
        ).encode()
        + _mca_well_select(row, col, labware)
        + b'",0,0);'
    )

    return command


def mca_dispense(
    grid_site: GridSite,
    row: int,
    col: int,
    volume: int | float,
    liquid_class: str,
    labware: Labware | str = "greiner96",
):
    """Advanced dispense command."""

    if isinstance(labware, str):
        labware = labwares[labware]

    command = (
        (
            "B;MCADispense("
            f'"{liquid_class}",'
            f'"{volume:g}",'
            f"{grid_site.grid},"
            f'{grid_site.site},"'
        ).encode()
        + _mca_well_select(row, col, labware)
        + b'",0,0);'
    )

    return command


def mca_mix(
    grid_site: GridSite,
    row: int,
    col: int,
    volume: int | float,
    liquid_class: str,
    labware: Labware | str = "greiner96",
):
    """Mix dispense command."""

    if isinstance(labware, str):
        labware = labwares[labware]

    command = (
        (
            "B;MCAMix("
            f'"{liquid_class}",'
            f'"{volume:g}",'
            f"{grid_site.grid},"
            f'{grid_site.site},"'
        ).encode()
        + _mca_well_select(row, col, labware)
        + b'",0,0);'
    )

    return command


def mca_get_tips(grid_site: GridSite, airgap: int = 20):
    """Get tips for mca."""
    return f"B;MCAGetDitis({grid_site.grid},{grid_site.site},{airgap},96,0,0);".encode()


def mca_drop_tips(grid_site: GridSite, waste_site=0):
    """Drop tips for mca."""
    return f"B;MCADropDitis({grid_site.grid},{grid_site.site},1,{waste_site},0,0);".encode()


def _mca_well_select(row: int, col: int, labware: Labware):
    """Generate well select string for mca."""
    row = (row - 1) % 2
    col = (col - 1) % 2

    def _encode_mca_well_select(well, enc=7):
        def _use_well(_well):
            return (labware.spacing == 1) or (
                (_well // labware.rows) % 2 == col and (_well % labware.rows) % 2 == row
            )

        encoder = 2 ** np.arange(enc) * np.array(
            list(map(_use_well, np.arange(enc) + well))
        )
        return int(sum(encoder) + 48).to_bytes(1, "big") if sum(encoder) > 0 else b"0"

    nwells = labware.rows * labware.cols
    sequence = [
        f"{labware.cols:02x}{labware.rows:02x}".encode(),
        *map(_encode_mca_well_select, np.arange(0, nwells - 6, step=7)),
        _encode_mca_well_select(nwells - nwells % 7, enc=nwells % 7),
    ]
    return b"".join(sequence)
